{$M 4096,0,655360}
{$G-,F-,S-,R-}

{{$DEFINE DEBUG}

{$IFDEF DEBUG}
{$S+,R+}
{$ENDIF}

{{$DEFINE PROFILE} {turns on "visual profiling" -- democoders know what I mean}
Program _8088flex;

{
A "flexible" version of the 8088 Corruption player.  JL20070218.
Completely rewritten to use a ring buffer and sane message passing
on 20120925.
Please see theidea.txt for the basic idea on how the player works, programming
notes, etc.
}

uses
  m6845ctl,    {some CGA constants}
  objects,     {file stream support}
  support,     {basic support stuff I don't want to include CRT for}
  cmdlin,      {for parsing command-line options}
  flexglob,    {flex framework globals}
  tinterrupts, {8253 manipulation routines}
  dos,         {for getintvec/setintvec}
  soundlib;    {Sound Blaster detection/setup/init/teardown code}

const
  samplerate:word=22050;
  maxsamplerate=45454;   {this is the maximum the Soundblaster Pro/2.0 can do}
  intfps:byte=30;
  maxintfps=60;             {CGA CRT maxes out at 60Hz (slightly less, actually)}
  sndchunksize:word=0;   {size of one video frame's sound data}
  maxsndChunkSize=$4000; {a reasonal maximum to ensure we don't cross DMA
                         boundary}
  sndbufsize:word=0;     {size of actual sound buffer (2 chunks)}
  maxsndBufSize=maxsndChunkSize * 2;

  DataQueued:boolean=false; {If FALSE, then playback pauses while more
                            data is buffered from disk}
  realCGA:boolean=true;  {Running on real CGA? If not, assume VGA}

  cols=40; {okay, so it's not THAT flexible... but I have no plans to support}
  rows=25; {anything other than 40x25 at this time}
  scrDataSize=cols*rows*2;

  avPayloadSize:word=0;    {size of a videodata+audiodata payload}

  numPauses:word=0;
  mono_output:boolean=false;
  quiet:boolean=false;
  maxRingBuffers=256;
  framecount:word=0;
  IOmult:byte=4;
  IOmask:byte=3;
  IOshift:byte=7;

  {to calculate CPU "utilization" of the player (really inverse of idle)}
  cpuUsageTicks:longint=0;
  cpuUsageCounter:longint=0;
  cpuUsageCounterIdle:longint=0;
  IOUsageTicks:longint=0;

  Int1CSave:pointer=nil;

type
  tscreendata=array[0..scrDataSize-1] of byte;
  pscreendata=^tscreendata;

  {Advantage of using [0..maxsndchunksize-1] is that we don`t have to hardcode
  params any more. Disadvantage?  Can't use new/dispose any more; we have to
  manually keep track of how large our buffers are.}
  tsnddata=array[0..maxsndChunkSize-1] of byte;
  psnddata=^tsnddata;

  tframedata=record
    vidbuf:tscreendata;
    sndbuf:tsnddata;
  end;
  pframedata=^tframedata;

  {Ring buffer object.  Allocates buffers of size BUFSIZE until doing so
  would drop free memory below MEMRESERVE.  Returns pointers to buffers,
  sequentially, wrapping around the ring.
  Concept is that the HEAD of the ring moves forward to grow the ring,
  and the TAIL is what moves forward to shrink the ring.  (Think of a snake.)
  New data is "fed" to the head, which grows.  Data is "consumed" from
  the tail, which shrinks toward the head.
  When the tail and head are equal, the ring buffer is empty.

  Buffers are filled by checking if not ring.isFull,
  then storing in buffers[bufferHead], then calling GrowHead.
  Buffers are consumed by checking if not ring.isEmpty,
  then using buffers[bufferTail], then calling ShrinkTail.

  Because the ring buffer routines are called from main code as well as
  interrupt code, certain critical functions are "wrapped" in the main
  code to ensure atomic operation.

  }
  ringbuffer=object(TObject)
    buffers:array[0..maxRingBuffers-1] of pointer;
    bufferMax,
    bufferHead,
    bufferTail,
    buffersInUse,
    buffersize:word;
    Constructor Init(bufsize:word;memreserve:longint);
    Destructor Done; virtual;
    Procedure GrowHead; {feed data into the ring}
    Procedure ShrinkHead; {adjust head backward if necessary (ie. aborted action)}
    Procedure ShrinkTail; {consume data from the ring}
    Function isEmpty:boolean;
    Function isFull:boolean;
  end;

var
  f:tDosStream;
  header:avheadrec;
  playp:pointer;
  AVRing:ringbuffer;
  {TicksSinceMidnight:longint absolute $0040:$006c;}
  vidp:pScreenData;
  {tmpFrameHandle:pframedata;}
  fname:string;

Constructor ringbuffer.Init(bufsize:word;memreserve:longint);
{We must keep buffermax one step below its maximum value so that
bufferhead has somewhere to step into to indicate out of bounds.}
begin
  Inherited Init;

  bufferhead:=0; buffertail:=0; buffermax:=0; buffersInUse:=0;
  buffersize:=bufsize;
  while (maxavail > (buffersize+memreserve)) and (buffermax < maxRingBuffers-1) do begin
    getmem(buffers[buffermax],buffersize);
    inc(buffermax);
  end;
  dec(bufferMax); {adjust bufferMax to point to the last successful allocation}
end;

Destructor ringbuffer.Done;
var
  w:word;
begin
  for w:=0 to buffermax do freemem(buffers[w],buffersize);
  buffermax:=0; bufferhead:=0; buffertail:=0; buffersize:=0; buffersInUse:=0;

  Inherited Done;
end;

Procedure ringbuffer.GrowHead; {feed data into the ring}
begin
  if not isFull then begin
    inc(bufferHead);
    if bufferHead>bufferMax then bufferHead:=0; {wrap around to 0}
    inc(buffersInUse);
  end;
end;

Procedure ringbuffer.ShrinkHead; {adjust head backward if necessary (ie. aborted action)}
begin
  if not isEmpty then begin
    dec(bufferHead); {roll back our Head}
    if bufferHead>bufferMax then bufferHead:=bufferMax; {handle wraparound past 0}
    dec(buffersInUse);
  end;
end;

Procedure ringbuffer.ShrinkTail; {consume data from the ring}
begin
  if not isEmpty then begin
    inc(bufferTail);
    if bufferTail>bufferMax then bufferTail:=0; {wrap around}
    dec(buffersInUse);
  end;
end;

Function ringbuffer.isEmpty:boolean;
begin
  isEmpty:=(bufferTail=bufferHead);
end;

Function ringbuffer.isFull:boolean;
begin
  isFull:=(buffersInUse=bufferMax);
end;

(*Function ringbuffer.isFull:boolean;
var
  b,found:byte;
begin
  found:=0; b:=bufferTail;
  repeat
    inc(found);
    inc(b); if b>bufferMax then b:=0; {handle wraparound}
  until b=bufferHead;
  isFull:=(found=bufferMax); {true if head-tail = total size of queue}
end;*)

procedure userTickHandler; interrupt;
begin
  inc(cpuUsageTicks);
end;

Procedure PrintHelp;
begin
  asm
    push ds
    jmp @start
@message:
    db 'Usage:  8088flex <switches> [filename] <switches>',0dh,0ah
    db '  <filename> is the video to play',0dh,0ah
    db '  /? or /h   This help',0dh,0ah
    db '  /q         Quiet -- do not attempt to use a sound card',0dh,0ah
    db '  /xN        I/O Multiplication factor (default 4, range 1-4)',0dh,0ah
    db '  /m         Force ugly output on monochrome screen',0dh,0ah
{    db '  /sx.x      Speed multiplier (default 1.0)',0dh,0ah}
    db '$'
@start:
    mov ax,0900h
    lea dx,@message
    mov bx,cs
    mov ds,bx
    int 21h
    pop ds
  end;
  halt(255);
end;

Procedure prepPlayer;
var
  l:longint;
begin
  {handle user input}
  if (paramcount=0) or is_Param('h') or is_Param('?') then printHelp;
  if is_param('q') then begin
    quiet:=true;
    writeln('User requested silence; no sound card initialization will take place.');
  end;
  if is_param('x') then IOMult:=param_int('x');
  if IOMult<1 then IOMult:=1; if IOMult>4 then IOMult:=4;
  IOShift:=IOMult-1;
  IOMask:=(1 SHL IOShift)-1;
  if non_flag_count>0 then fname:=non_flag_param(1) else fname:='8088_cor.dat';
  if not fileexists(fname) then fatalerror(1,'Filename "'+fname+'" not found');

  f.init(fname,stOpenRead);
  if f.status<>stOK then fatalerror(2,'There was a problem opening the media file.');
  {read header}
  f.read(header,sizeof(header));
  if header.fileheader[0]<>'T' {header should start with TMAV}
    then fatalerror(127,'Header signature doesn''t indicate TMAV file.');

  sndchunksize:=header.chunksize;
  samplerate:=header.samplerate;
  sndbufsize:=sndchunksize*2; {because buffer is actually divided into two IRQ firings}
  If (Mem[$0000:$0449]=7) or is_param('m') then begin
    mono_output:=true;
    vidp:=ptr($b000,0)
  end else begin
    mono_output:=false;
    vidp:=ptr($b800,0);
  end;
  avPayloadSize:=scrDataSize+sndChunkSize;
  if (header.features AND f_padding)=1 then begin {if version 1 then we pad to sector boundaries}
    {padding at the speed of light}
    asm
      add AVPayloadSize,(PadBoundary-1)
      and AVPayloadSize,NOT (PadBoundary-1)
    end;
    f.seek(512);
    writeln('This file''s AV chunk data is aligned to ',PadBoundary,'-byte boundaries.');
  end;

  writeln('Video appears to be ',samplerate / sndchunksize:2:2,'fps with an audio sample rate of ',samplerate,'.');
  intfps:=round(samplerate/sndchunksize);
  if intfps<4 then intfps:=4; if intfps>maxintfps then intfps:=maxintfps;
  {init the ring buffer}
  l:=memavail;
  AVRing.init(avPayloadSize * (1 SHL IOShift),4096); {leave at least a block of 4K heap available}
  writeln('Ring buffer established with ',
    AVRing.buffermax,'x',AVRing.buffersize,'-byte segments using ',
    (l-memavail) div 1024,' KiB (',maxavail div 1024,' KiB free)');
  {setup and determine cpu idle counter rate}
  GetIntVec($1C,Int1CSave);
  SetIntVec($1C,@userTickHandler);

  {see how high counter goes for 1 tick when idle}
  write('Calibrating CPU idle counter... ');
  l:=cpuUsageTicks;
  repeat
    asm
      hlt
    end
  until l<>cpuUsageTicks; {wait until beginning of tick}
  l:=cpuUsageTicks;
  repeat
    inc(cpuUsageCounterIdle);
  until cpuUsageTicks >= l+8;
  cpuUsageCounterIdle:=cpuUsageCounterIdle div 8;
  writeln('Done.');
end;

Procedure setupVideoMode;

  Procedure colorCal;
  {displays color pattern to calibrate monitor}

  const
    length=5;
    labels:array[0..3] of string[40]=(
      ' blk  blu green cyn  red magntbrown gry ',
      'dark lightlightlightlightlight',
      ' gry  blu green cyn  red magnt yel white',
      'Calibrate monitor, then hit Enter:'
    );
  var
    y,x,baz:byte;

    procedure putcharxy(x,y,ch:byte);
    begin
      mem[$b800:(y*80)+(x*2)]:=ch;
    end;

    procedure putAttrXY(x,y,fore,back:byte);
    begin
      mem[$b800:(y*80)+(x*2)+1]:=(back SHL 4) or fore;
    end;

  begin
    for y:=0 to 11 do begin
      for x:=0 to 7 do begin
        for baz:=0 to length-1 do begin
          {top bars}
          putCharXY((x*length)+baz,y,   32);
          putAttrXY((x*length)+baz,y,   15,x);
          {bottom bars}
          putCharXY((x*length)+baz,y+12,32);
          putAttrXY((x*length)+baz,y+12,7, x+8);
        end;
      end;
    end;
    {write out the labels}
    for x:=0 to 39 do begin
      putCharXY(x,6,byte(labels[0][x+1]));
      putCharXY(x,17,byte(labels[1][x+1]));
      putCharXY(x,18,byte(labels[2][x+1]));
      putCharXY(x,24,byte(labels[3][x+1]));
    end;
  end;

var
  ch:char;

begin
  if not mono_output then begin
    writeln('Are you running this program on a *real* hardware CGA card? (Y/N)');
    writeln('(If running on an emulator, try answering "N") ');
    ch:=upcase(readkeychar);
    if ch='Y' then realCGA:=true else realCGA:=false;

    {set up the video mode}
    if realCGA
      then asm
        mov  ax,01h {set 40x25}
        int  10h
        mov  dx,3d8h
        mov  al,c_videosignal_enable {blinking not included, so blinking is off}
        out  dx,al
      end else asm {if VGA, do something completely different}
        MOV AX,1200h {set up 200 scan lines}
        MOV BL,30h
        INT 10h
        MOV AX,0001h {set up normal 40x25 text mode}
        INT 10h
        MOV AX,1102h {load 80x50 vga character set}
        MOV BL,00
        INT 10h
        mov ax,1003h {disable blink on ega/vga}
        mov bl,0
        int 10h
      end;

    colorcal;
    if keypressed then repeat readkeychar until not keypressed;
    readkeychar;
    CursorOFF(0);
  end else begin
    writeln('This demonstration was developed for CGA color 40x25 mode.');
    writeln('Monochrome emulation looks truly horrible and is slow.');
    write('Are you sure you want to try running this for Monochrome? (Y/N) ');
    ch:=upcase(readkeychar);
    if ch<>'Y' then begin
      writeln('Wise choice!');
      halt(1);
    end;
    {set up the video mode}
    asm
      mov  ax,07h {mono 80x25}
      int  10h
    end;
    CursorOFF(0);
  end;
  (*if not mono_output
    then fillchar(vidp^,scrDataSize,0)
    else fillchar(vidp^,scrDataSize*2,$70);*)
  {if mono_output then fillchar(vidp^,scrDataSize*2,$70);}
  if mono_output then asm
    les di,vidp
    mov cx,scrDataSize
    shl cx,1
    mov ax,$70db {db = 219, solid char}
    rep stosw
  end;
end;

procedure soundIntCallerCGA; far;
{
This is where the magic happens.  If we are executing in this code, it means
the sound card just got *finished* playing a buffer, which is our cue to
prepare the next frame.

No check for re-entrancy is performed in this particular snippet because
the parent code in the soundblaster and interrupt routines do that for us.
(In case it wasn't obvious, this is NOT an interrupt handler procedure!)
}

var
  sndbufp:pointer;    {temp pointer to where we should copy our new audio data}
  framehp:pointer;    {temp pointer video data comes from}
  localFrameHandle:pFrameData;               {local handle for our av chunk}

begin
  {$IFDEF PROFILE} port[$3d9]:=1; {$ENDIF}
  if not quiet then begin
    sndbufp:=play.p; {point to dma buffer}
    if play.activeblock = 2       {If in one half, advance to the second half}
      then inc(word(sndbufp),sndChunkSize); {can do this because little-endian}
  end;
  if DataQueued    {if we're allowed to pull AV chunks from the FIFO queue:}
    then begin
      {get the handle to the data}
      with AVRing do localFrameHandle:=buffers[bufferTail];
      {advance buffer pointer based on frame count if necessary}
      inc(word(localFrameHandle),avPayloadSize * (framecount AND IOMask));

      {$IFDEF PROFILE} port[$3d9]:=2; {$ENDIF}
      if not quiet then begin
        framehp:=@localFrameHandle^.sndbuf;
        asm
          mov  dx,ds
          les  di,sndbufp
          cld
          mov  cx,sndchunksize
          lds  si,framehp {don't need vars from DS any more so this is fine}
          shr  cx,1       {if odd, bit will be shifted off into carry flag}
          rep  movsw
          adc  cx,cx      {if carry set, cx=1 else cx=0}
          rep  movsb      {if cx=0 nothing happens. We do this to avoid JMPing}
          mov  ds,dx
        end;
      end;
      {$IFDEF PROFILE} port[$3d9]:=3; {$ENDIF}
      framehp:=@localFrameHandle^.vidbuf;
      asm
        mov  dx,ds
        les  di,vidp
        cld
        mov  cx,scrDataSize
        lds  si,framehp {don't need vars from DS any more so this is fine}
        shr  cx,1
        rep  movsw
        {no need to handle odd because frame data always even-sized}
        mov  ds,dx
      end;
      {we're done with the data, NOW shrink the tail}
      if (framecount AND IOMask)=IOMask then AVRing.ShrinkTail;
      inc(framecount);
      {$IFDEF PROFILE} port[$3d9]:=4; {$ENDIF}

    end else if not quiet then begin   {if we are NOT allowed to pull AV from the FIFO queue:}
      asm {fill buffer with silence if we have nothing to put in it}
        les  di,sndbufp
        cld
        mov  cx,sndChunkSize
        shr  cx,1
        mov  ax,$7f7f
        rep  stosw
        adc  cx,cx     {if carry set, cx=1 else cx=0}
        rep  stosb     {if cx=0 nothing happens.  Avoids JMP, which is slow.}
      end;
    end;
  {toggle sound double-buffer blocks}
  {if play.activeblock=1 then play.activeblock:=2 else play.activeblock:=1;}
  asm
    xor play.activeblock,00000011b {toggles between 1 and 2}
  end;
  {$IFDEF PROFILE} port[$3d9]:=0; {$ENDIF}
end;

procedure soundIntCallerMDA; far;

var
  sndbufp:pointer;    {temp pointer to where we should copy our new audio data}
  framehp:pointer;    {temp pointer video data comes from}
  localFrameHandle:pFrameData;               {local handle for our av chunk}

begin
  if not quiet then begin
    sndbufp:=play.p; {point to dma buffer}
    if play.activeblock = 2       {If in one half, advance to the second half}
      then inc(word(sndbufp),sndChunkSize); {can do this because little-endian}
  end;
  if DataQueued    {if we're allowed to pull AV chunks from the FIFO queue:}
    then begin
      {get the handle to the data}
      with AVRing do localFrameHandle:=buffers[bufferTail];
      {advance buffer pointer based on frame count if necessary}
      inc(word(localFrameHandle),avPayloadSize * (framecount AND IOMask));

      if not quiet then begin
        framehp:=@localFrameHandle^.sndbuf;
        asm
          mov  dx,ds
          les  di,sndbufp
          cld
          mov  cx,sndchunksize
          lds  si,framehp {don't need vars from DS any more so this is fine}
          shr  cx,1       {if odd, bit will be shifted off into carry flag}
          rep  movsw
          adc  cx,cx      {if carry set, cx=1 else cx=0}
          rep  movsb      {if cx=0 nothing happens. We do this to avoid JMPing}
          mov  ds,dx
        end;
      end;
      framehp:=@localFrameHandle^.vidbuf;
      asm
        mov  dx,ds
        les  di,vidp
        add  di,cols    {center the image}
        cld  {put cld here so it can be prefetched during the slow LES opcode}
        mov  bx,rows
        lds  si,framehp {don't need vars from DS any more so this is fine}
        mov  ah,$70 {this is a monochrome attribute fixup if we decide to use it}
      @dorow:
        mov  cx,cols / 4
      @loopit:
        lodsb   {load character from source}
        inc si  {skip over attribute byte in source}
        stosb   {store character to destination}
        inc di  {skip over attribute byte in destination}

        lodsb
        inc si
        stosb
        inc di

        lodsb
        inc si
        stosb
        inc di

        lodsb
        inc si
        stosb
        inc di

        loop @loopit
        add  di,cols * 2
        dec  bx
        cmp  bx,0
        jne  @dorow

        mov  ds,dx
      end;
      {we're done with the data, NOW shrink the tail}
      if (framecount AND IOMask)=IOMask then AVRing.ShrinkTail;
      inc(framecount);

    end else if not quiet then begin   {if we are NOT allowed to pull AV from the FIFO queue:}
      asm {fill buffer with silence if we have nothing to put in it}
        les  di,sndbufp
        cld
        mov  cx,sndChunkSize
        shr  cx,1
        mov  ax,$7f7f
        rep  stosw
        adc  cx,cx     {if carry set, cx=1 else cx=0}
        rep  stosb     {if cx=0 nothing happens.  Avoids JMP, which is slow.}
      end;
    end;
  {toggle sound double-buffer blocks}
  {if play.activeblock=1 then play.activeblock:=2 else play.activeblock:=1;}
  asm
    xor play.activeblock,00000011b {toggles between 1 and 2}
  end;
end;

procedure enqueuefile;
var
  oldticks:longint;
begin
  {clear fat-finger typing so player doesn't immediately exit}
  if keypressed then repeat readkeychar until not keypressed;
  {must be atomic; 32-bit, 2 instructions, so we wrap it}
  asm pushf; cli end; cpuUsageTicks:=0; asm popf end; cpuUsageCounter:=0;
  with AVRing do begin
    repeat
      if not isFull then begin {if we have space for video+audio data and some extra for handle}
        oldticks:=cpuUsageTicks;
        f.read(buffers[bufferHead]^,avPayloadSize * (1 SHL IOShift));
        inc(IOUsageTicks,cpuUsageTicks-oldticks);
        GrowHead; {mark buffer segment as claimed}
        if f.status=stOk then begin
          {If we dip below a handful of buffer segments, or we have SUCH a
          slow harddisk that we consumed all our data without noticing
          leaving the ring buffer empty, PAUSE PLAYBACK to stop consuming}
          if (buffersInUse < 3) or isEmpty then begin
            if dataqueued then inc(numPauses);
            dataqueued:=false; {this must be atomic or else playback
            interrupt might try to play non-existent data.  Luckily, it
            compiles to a single instruction.}
          end;
        end else begin {if disk was not read ok (error, EOF, etc.) then backpeddle}
          ShrinkHead;
        end;
      end else begin
        DataQueued:=true; {ring buffer is full, we're ready to play}
      end;
      inc(cpuUsageCounter); {record idle time}
      if keypressed then exit;
    until (f.status<>stOK); {until we reach end-of-file or other error}
  end;
  f.done;
  {If entire file fit in RAM, play it. Don't leave here without playback enabled!}
  DataQueued:=true;
end;

Procedure primeSoundcard;
var
  smult:real;
begin
  if initblaster <> 0 then begin
    SetIntVec($01C,Int1CSave); {stop counting ticks}
    fatalError(4,'Couldn''t initialize Sound Blaster');
  end;
  setmemsize(sndbufsize);
  {$IFDEF DPMI} {someday we might take advantage of 286 protected mode ram}
  playp := ptr(play.p,0);
  {$ELSE}
  playp := play.p;
  {$ENDIF}

  fillchar(playp^,sndbufsize,127); {init buffer with silence}
  if not mono_output
    then play.callproc := soundIntCallerCGA
    else play.callproc := soundIntCallerMDA;
  if is_param('s') then begin
    smult:=param_real('s');
    if trunc(samplerate * smult) < maxsamplerate
      then samplerate:=trunc(samplerate * smult)
      else samplerate:=maxsamplerate;
  end;
  initplay(0,play.p,sndbufsize,samplerate,snd_8bit+snd_mono+snd_nonsigned);
  if not mono_output then writeln(getblasterstring,' initialized.');
end;

{$F+}
procedure noSoundIntCaller; Interrupt;

begin
  play.callproc;

  {We want to be nice, and will maintain the BIOS interrupt}
  inc(PITCycles,Chan0Counter); {Keep track of how many PIT cycles have gone by}
  if longrec(PITCycles).hi <> 0 then begin {Did we roll over?  Is it time to call the 18.2Hz BIOS handler?}
    longrec(PITCycles).hi:=0; {Update our PIT cycles counter}
    asm pushf end; {simulate an interrupt by pushing flags, then CALLing handler}
    BIOSTimerHandler; {this will acknowledge the interrupt}
  end
  else
    Port[$20] := $20; {send EndOfInterrupt to the PIC to ackn. the interrupt}
end;
{$F-}

procedure primeNoCard; {silence!  We will use the system timer!}
begin
  if intfps<20 then intfps:=20; {timer can only go as low as 18.2}
  if not mono_output
    then play.callproc := soundIntCallerCGA
    else play.callproc := soundIntCallerMDA;
  SetTimerHz(@noSoundIntCaller, intfps); {set to fire at video's rough framerate}
end;

Procedure shutItDown;
begin
  SetIntVec($01C,Int1CSave); {stop counting ticks}
  if quiet
    then CleanUpTimer {stop our internal "sound" handler}
    else exitplay; {shutdown Sound Blaster}
  CursorON;

  if not mono_output
    then begin
      asm
        mov ax,3
        int 10h
      end;
    end else begin
      asm
        mov ax,7
        int 10h
      end;
    end;
  asm
    push ds
    jmp @start
@message:
    db '8088 Corruption Player v0.7, 20120920',0dh,0ah
    db 'Idea, fast encoder, and player engine by Trixter / Hornet',0dh,0ah
    db 'Quality encoder by gODjR / iq2000',0dh,0ah
    db 'Sound Blaster shell (SB "2.01" ;-) by Stefan Goehler',0dh,0ah
    db 'Thanks to Klaus Hartnegg and Kris Heidenstrom for interrupt advice',0dh,0ah,0ah
    db '$'
@start:
    mov ax,0900h
    lea dx,@message
    mov bx,cs
    mov ds,bx
    int 21h
    pop ds
  end;
  writeln('Playback statistics:');
  writeln('# of pauses during playback: ',numPauses);
  {$IFDEF DEBUG}
  writeln(
  'cpuusagecounter: ',cpuusagecounter,
  ' cpuusageticks: ',cpuusageticks,
  ' cpuusagecounteridle: ',cpuusagecounteridle
  );
  {$ENDIF}
  writeln('CPU was ',(100*(cpuUsageCounter div cpuUsageTicks)) div cpuUsageCounterIdle,'% idle');
  writeln((100*IOUsageTicks) div cpuUsageTicks,'% of total playback time was spent reading from disk');
  writeln('# of buffer segments still active: ',AVRing.buffersInUse);

  AVRing.done;
  if keypressed then repeat readkeychar until not keypressed; {absorb keystroke that was used to exit playback}
end;

begin
  prepPlayer;
  setupVideoMode;
  if quiet
    then primeNoCard
    else primeSoundcard;
  if not mono_output then writeln('Performing initial buffering...');
  EnqueueFile;
  repeat
    {do nothing, because if we're here, we've loaded everything into RAM
    and are just waiting for everything to be played}
    inc(cpuUsageCounter); {record idle time}
  (*until frameHandles.isempty or not dataqueued;*)
  until AVRing.isempty or not dataqueued;
  shutItDown;
end.
