program avcomp;
{
Compiles a TMV file from the source converted frames and raw audio dump

Header of a TMV file:

'TMAV'     - 4 bytes, header signature
samplerate - 2 bytes, word value, playback sample rate
chunksize  - 2 bytes, word value, audio chunk size per frame
compmethod - 1 byte, specifies the compression method (currently only 0)
charcols   - 1 byte, number of text chars across (currently always 40)
charrows   - 1 byte, number of text chars down (currently always 25)
version#   - 1 byte, version number, used for special handling:
  0: no special handling
  1: All AV chunks are padded to the next 512-byte boundary

Notes:

Framerate is not explicitly in the file; you calc (samplerate/chunksize)
to get it.

If audiochunk is not an integer, there will be drift!  So the converter
will suggest closest match that prevents drift if that arises.  Closest
is found by taking the calc'd chunksize, trunc' and round'
it, multiplying by user-entered framerate, and choosing the result
closest to the original samplerate.
}

uses
  support,
  objects,
  flexglob;

const
  maxBufSize:word=65536-16;

var
  outstream:pBufStream;
  framebuf:pointer; framesize:word;
  finf,foutf,inaudf:file;
  numread:word;
  tempfname,soundfname,outfname,basename:string;
  header:avheadrec;
  startfn,endfn,curframe:word; {starting and ending frame number}
  framerate:real;
  temps:string;
  tempt,tempr,distt,distr:real;

function askforstring(s:string):string;
var
  temps:string;
begin
  write(s,' '); readln(temps); askforstring:=temps;
end;

function askforreal(s:string):real;
var
  tempr:real;
begin
  write(s,' '); readln(tempr); askforreal:=tempr;
end;

function askforword(s:string):word;
var
  tempw:word;
begin
  write(s,' '); readln(tempw); askforword:=tempw;
end;

Procedure printparms;
begin
  writeln('Current selections: ');
  writeln('samplerate: ',header.samplerate);
  writeln('chunk size: ',header.chunksize);
  writeln(' framerate: ',framerate:2:2);
  writeln('  basename: ',basename);
  writeln('startframe: ',startfn);
  writeln('  endframe: ',endfn);
  writeln('audiofname: ',soundfname);
end;

begin
  {init header vals}
  with header do begin
    move(headsig,fileheader,4);
    compmethod:=0; charcols:=40; charrows:=25; features:=0;
    framesize:=charcols*charrows*2;
  end;
  getmem(framebuf,maxBufSize);

  {get user input}
  basename:=askforstring('Base filename of frames?');
  outfname:=basename+'.tmv';
  soundfname:=askforstring('8-bit PCM audio filename?');
  startfn:=askforword('Staring frame number?');
  endfn:=askforword('Ending frame number?');
  header.samplerate:=askforword('Audio playback sample rate?');
  framerate:=askforreal('Framerate of the source movie? (fractions like 23.976 are ok)');

  with header do begin
    {check for fractional chunk size, can't have that}
    if frac(samplerate / framerate)<>0 then begin
      writeln('A samplerate of ',samplerate,' divided by a framerate of ',framerate:2:2);
      writeln('would result in an audio chunk size of ',samplerate/framerate:2:2);
      write('so the samplerate needs to be adjusted.  Adjusting to ');
      chunksize:=round(samplerate/framerate);
      {tempr:=chunksize * framerate;}
      samplerate:=round(chunksize * framerate);
      writeln(samplerate,'.');
    end else begin
      chunksize:=trunc(samplerate / framerate);
    end;
    writeln('This gives us a chunk size of ',chunksize,'.');
  end;

  printparms;
  temps:=askforstring('Hit enter to begin, or type quit to abort');
  if temps<>'' then fatalerror(1,'user abort');

  {convert}
  assign(inaudf,soundfname);
  reset(inaudf,1);
  new(outStream);
  outStream^.init(outfname,stCreate,maxBufSize);
  outStream^.write(header,sizeof(header));
  for curframe:=startfn to endfn do begin
    tempfname:=basename+inttostr(curframe)+'.dmp'; assign(finf,tempfname); reset(finf,1);
    blockread(finf,framebuf^,framesize);
    outstream^.write(framebuf^,framesize);
    {common to run out of audio on the last frame due to rounding, so handle it}
    fillchar(framebuf^,127,header.chunksize);
    blockread(inaudf,framebuf^,header.chunksize,numread);
    outStream^.write(framebuf^,header.chunksize);
    close(finf);
    if numread<>header.chunksize then begin
      writeln('Ran out of audio data;');
      if curframe=endfn
        then writeln('we''re on the last frame, so this is ok.')
        else writeln('we''re on frame ',curframe,' so something is wrong.');
      break;
    end;
    if curframe and $0f=0 {status updates}
      then write(#13,curframe);
  end;
  close(inaudf);
  outStream^.done;
  writeln(#13,endfn-startfn,' A/V chunks written to ',outfname);
end.
